/*
	ProWesS, PROGS Windowing System

	PWRemove

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 14 January 1995
	03.04.1996 bug fixed which prevented proper removal of sleep object
			   in remove_region, some code read
				   else if (system->regions=region)
			   instead of
				   else if (system->regions==region)
	11.05.1996 PWRemove now safe when NULL passed as object
*/

#include <atomic.h>
#include <mem.h>

#include "private.h"
#include "PWHandler.h"

/* forward declarations */
static Error remove_region(PWSystem *system, PWRegion *region);
static Error remove_keypress(PWSystem *system, PWKeyPress *keypress);

Error PWRemove(PWObject object)
{
	Error err=ERR_OK;
	PWSystem *system;

	if (object==NULL) return err;

	system=object->event.region->system;

	switch (object->type->event_type)
	{
	case PW_TYPE_EVENT_REGION:
		if (system->activated)
		{
			system->status|=SSTATUS_WINCHANGE|SSTATUS_REMOVE;
			object->status|=OSTATUS_REMOVE;
		}
		else
			err=remove_region(system,object->event.region);
		break;
	case PW_TYPE_EVENT_KEYPRESS:
		err=remove_keypress(system,object->event.keypress);
		break;
	}
	if (err) return err;

	/* check if there is still an object in the system, or discard system */
	if (system->regions==NULL)
	{
		if (system->sleep) remove_region(system,system->sleep);
		while (system->keypresses) remove_keypress(system,system->keypresses);
		while (system->state)
		{
			PWDummyTypeState *state=system->state;
			system->state=state->next;
			MEMRelease(state);
		}
		PWClose(system);
		PFGstateRemove(system->gstate);
		MEMHeapZap(system->heap);
	}
	return err;
}

int RegionRemoved(PWRegion *region)
{
	if (region->owner && (region->owner->status&OSTATUS_REMOVE))
	{
		remove_region(region->system,region);
		return TRUE;
	}
	else
	{
		for (region=region->children ; region ; region=region->next)
			if (RegionRemoved(region)) return TRUE;
		return FALSE;
	}
}

Error PWRemoveIndicated(PWSystem *system)
{
	PWRegion *region;
	system->status&=~SSTATUS_REMOVE;
	while (TRUE)
	{
		for (region=system->regions ; region ; region=region->next)
			if (RegionRemoved(region)) continue;
		break;
	}
}

static Error remove_region(PWSystem *system, PWRegion *region)
{
	PWRegion **trail;
	Error err;

	/* start by removing the owner object - if exists */
	if (region->owner)
	{
		if (region->owner==system->catchkeys)
		{
			if (region->owner->type->PWHandle)
				region->owner->type->PWHandle(region->owner,PW_EVENT_CATCHEND);
			system->catchkeys=NULL;
		}
		err=region->owner->type->PWRemove(region->owner);
		if (err) return err;
	}

	/* now remove all the children */
	while (region->children)
		remove_region(system,region->children);

	/* unlink region from system */
	if (region->prev)
	{
		region->prev->next=region->next;
	}
	else
	{
		if (region->parent)
		{
			region->parent->children=region->next;
		}
		else if (system->regions==region)
			system->regions=region->next;
		else
			system->sleep=NULL;
	}
	if (region->next)
		region->next->prev=region->prev;

	/* release region */
	MEMHeapRelease(system->heap,region);

	/* make sure the window is rebuilt if any */
	if (system->regions!=NULL)
		system->status|=SSTATUS_WINCHANGE;

	return ERR_OK;
}

static Error remove_keypress(PWSystem *system, PWKeyPress *keypress)
{
	PWKeyPress **trail;
	Error err;

	/* start by removing the owner object - if exists */
	if (keypress->owner)
	{
		err=keypress->owner->type->PWRemove(keypress->owner);
		if (err) return err;
	}

	/* unlink keypress from system */
	trail=&system->keypresses;
	while (*trail!=keypress) trail=&(*trail)->next;
	*trail=keypress->next;

	/* release keypress */
	MEMHeapRelease(system->heap,keypress);

	return ERR_OK;
}
